
#include "detect.h"
#include "color.h"
#include "draw.h"
#include "math_approx.h"


// define values for visual debugging;
#define DEBUG_BALL  false
#define DEBUG_FIELD false
#define DRAW_FIELD  true
#define DRAW_BALL   true


// define min and max for integer
#define min(a, b) (((a) < (b)) ? (a) : (b))
#define max(a, b) (((a) > (b)) ? (a) : (b))


// user specific colors
uint8_t * color_ball  = red;
uint8_t * color_field = yellow;

uint16_t field_location_size[4] = {0, 0, 0, 0};     // left, right, top, bottom
uint8_t ball_match              = 55;               // in percent * 100; 55 = 55%

uint8_t  colComp_dark[]   = { 60,  5, 20, 255};
uint8_t  colComp_middle[] = { 90, 25, 25, 255};
uint8_t  colComp_bright[] = {160, 75, 75, 255};


/**
 * checks whether the amount of green is significant higher than other color channels
 * @param color the color
 * @param threshold the threshold for being green
 * @return true, if color is green
 */
inline bool isGreen(uint8_t * color, uint8_t threshold) {
    return ((color[1] - color[0]) > threshold &&
            (color[1] - color[2]) > threshold);
}

/**
 * checks whether the amount of red is higher than the other colors
 * @param color the color
 * @return true, if color is red enough
 */
inline bool isRed(uint8_t * color) {
    return (color[0] > (color[1] + color[2]) * 0.7);
}


/**
 * detects whether the given pixel could be part of the ball
 * @param pixel the pixel / color
 * @return true, if it could be part of the ball
 */
inline bool isPotencialPartOfBall(uint8_t * pixel) {
    return  (isRed(pixel) &&
                (colorDistanceRGB(pixel, colComp_middle) < 60 ||
                 colorDistanceRGB(pixel, colComp_dark)   < 25 ||
                 colorDistanceRGB(pixel, colComp_bright) < 25));
}

/**
 * detects whether the given pixel could be part of the field
 * @param pixel the pixel / color
 * @param pixel_next the pixel next to 'pixel', for comparison
 * @return true, if it could be part of the field
 */
inline static bool isPotencialBorderOfField(uint8_t * pixel, uint8_t * pixel_next) {
    return (colorDistanceRGB(pixel, black)       < 150 &&
            colorDistanceRGB(pixel, pixel_next)  <  25 &&
            !isGreen(pixel, 3));
}


/**
 * searches a given area for the ball.
 * @param srcPicture the image
 * @param center_y the row of the ball
 * @param center_x the column of the ball
 * @param radius the radius which will be searched for the ball
 * @param ball_location_size array of type uint16_t [4]
 */
void detectBall(pictureDataP srcPicture, uint16_t center_y, uint16_t center_x, uint8_t radius, uint16_t * ball_location_size) {
    uint32_t count      = 0;
    uint32_t area       = approx_square(2 * radius);

    for (int16_t row = -radius; row <= radius; row++)
    for (int16_t col = -radius; col <= radius; col++)
    // pythagoras, check whether the current pixelis inside the circle.
    if ((approx_squareroot(approx_square(abs(row)) + approx_square(abs(col)))) <= radius) {
        uint16_t pixel_row = 0;
        uint16_t pixel_col = 0;

        // check boundaries for the frame.
        if (center_y + row >= 0 && center_y + row < srcPicture->height)
            pixel_row = center_y + row;
        else if (center_y + row > srcPicture->height)
            pixel_row = srcPicture->height - 1;

        if (center_x + col >= 0 && center_x + col < srcPicture->width)
            pixel_col = center_x + col;
        else if (center_x + col > srcPicture->width)
            pixel_col = srcPicture->width - 1;

        uint8_t * pixel = &(srcPicture->rowPointers[pixel_row][pixel_col * 4]);
        if (isPotencialPartOfBall(pixel))
            count++;
    }

    // max value:       78%
    // area:            circle: pi * r**2 = pi * ((a**2) / 4)
    // square:          a**2
    // reduce fraction: (pi / 4) * 100 = 78.58% = 78%
    uint8_t match = count * 100 / area;

    // if the radius is greater than the last and the match
    // is between 56 and 78, we found the greater ball.
    if (match > 55 && radius >= ball_location_size[2]) {
        ball_location_size[0] = center_x;
        ball_location_size[1] = center_y;
        ball_location_size[2] = radius;
        ball_location_size[3] = match;
    }
}

/**
 * finds the potencial ball and draws it.
 * @param srcPicture the image
 * @param min_diameter the minimum diameter of the ball
 * @param max_diameter the maximum diameter of the ball
 * @return true, if ball has been found
 */
bool detectCircle(pictureDataP srcPicture, uint8_t min_diameter, uint8_t max_diameter) {
    // center_x, center_y, radius, match_percent
    uint16_t ball_location_size[] = {0, 0, 0, 0};

    for(uint16_t row = field_location_size[2]; row <= field_location_size[3]; row++)
    for(uint16_t col = field_location_size[0]; col <= field_location_size[1]; col++) {
        uint8_t * pixel = &srcPicture->rowPointers[row][col * 4];

        if (!DEBUG_FIELD)
            if (DEBUG_BALL) {
                if (isPotencialPartOfBall(pixel))
                    colorSetPixelRGBA(pixel, white);
                else
                    colorSetPixelRGBA(pixel, black);
            } else {
                if (isPotencialPartOfBall(pixel))
                    for (uint8_t diameter = min_diameter; diameter <= max_diameter * 2; diameter++)
                        detectBall(srcPicture, row + diameter/2, col + diameter/2, diameter/2, &ball_location_size[0]);
            }
    }


    // checks, if the found ball is inside the found field.
    if (DRAW_BALL &&
        ball_location_size[0] >= field_location_size[0] && ball_location_size[0] <= field_location_size[1] &&
        ball_location_size[1] >= field_location_size[2] && ball_location_size[1] <= field_location_size[3])
    {
        if (field_location_size[0] - ball_location_size[2] > 0 &&
            field_location_size[1] + ball_location_size[2] < srcPicture->width - 1 &&
            field_location_size[2] - ball_location_size[2] > 0 &&
            field_location_size[3] + ball_location_size[2] < srcPicture->height - 1)
            {
                drawCircle(srcPicture, color_ball, ball_location_size[0], ball_location_size[1], ball_location_size[2]);
            } else {
                drawLineHorizontal(srcPicture, color_ball, ball_location_size[1]);
                drawLineVertical(  srcPicture, color_ball, ball_location_size[0]);
            }

        return true;
    }

    return false;
}


/**
 * find potential rows being the border of the field
 * @param srcPicture the image
 */
void updateStatisticTB(pictureDataP srcPicture) {
    for (uint16_t row = 0; row < srcPicture->height; row++) {
        int16_t counter = 0;

        // check only the first and last quarter
        if (row == srcPicture->height / 4)
            row = srcPicture->height * 0.75;


        for (uint16_t col = 0; col < srcPicture->width - 1; col++) {
            uint8_t * pixel      = &srcPicture->rowPointers[row][ col      * 4];
            uint8_t * pixel_next = &srcPicture->rowPointers[row][(col + 1) * 4];

            if (isPotencialBorderOfField(pixel, pixel_next)) {
                counter++;

                if (DEBUG_FIELD)
                    colorSetPixelRGBA(pixel, black);
            } else {
                if (DEBUG_FIELD)
                    colorSetPixelRGBA(pixel, white);
            }
        }

        if(counter > srcPicture->width * 0.85)
            #pragma omp atomic
                statistical_border_tb_array[row]++;
    }
}

/**
 * find potential columns being the border of the field
 * @param srcPicture the image
 */
void updateStatisticLR(pictureDataP srcPicture) {
    for (uint16_t col = 0; col < srcPicture->width; col++) {
        int16_t counter = 0;

        // check only the first and last quarter
        if (col == srcPicture->width / 6)
            col = srcPicture->width * 0.83;


        for (uint16_t row = field_location_size[2]; row < field_location_size[3] - 1; row++) {
            uint8_t * pixel      = &srcPicture->rowPointers[row][ col      * 4];
            uint8_t * pixel_next = &srcPicture->rowPointers[row][(col + 1) * 4];

            if (isPotencialBorderOfField(pixel, pixel_next)) {
                counter++;

                if (DEBUG_FIELD)
                    colorSetPixelRGBA(pixel, black);
            } else {
                if (DEBUG_FIELD)
                    colorSetPixelRGBA(pixel, white);
            }
        }

        // the goal height is about 1/3 of the field height
        if( counter > (field_location_size[3] - field_location_size[2]) * 0.6 &&
            counter < (field_location_size[3] - field_location_size[2]) * 0.7)
                #pragma omp atomic
                    statistical_border_lr_array[col]++;
    }
}


/**
 * analyze the potential rows and get the border of the field
 * @param srcPicture the image
 */
void detectTableBorderTB(pictureDataP srcPicture) {
    // Looking for the upper boarder
    uint16_t upper_border       = 0;
    uint16_t upper_border_value = 0;
    for(uint16_t row = 0; row < srcPicture->height * 0.25; row++)
        if(statistical_border_tb_array[row] > upper_border_value) {
            upper_border_value = statistical_border_tb_array[row];
            upper_border       = row;
        }

    // Looking for the lower boarder
    uint16_t lower_border       = srcPicture->height - 1;
    uint16_t lower_border_value = 0;
    for(uint16_t row = srcPicture->height * 0.75; row < srcPicture->height; row++)
        if(statistical_border_tb_array[row] > lower_border_value) {
            lower_border_value = statistical_border_tb_array[row];
            lower_border       = row;
        }


    // update location and size of field
    field_location_size[2] = upper_border;               // top
    field_location_size[3] = lower_border;               // bottom
}

/**
 * analyze the potential columns and get the border of the field
 * @param srcPicture the image
 */
void detectTableBorderLR(pictureDataP srcPicture) {
    // search for the left border
    uint16_t left_border       = 1;
    uint16_t left_border_value = 0;
    for (uint16_t col = 1; col < srcPicture->width / 6; col++)
        if(statistical_border_lr_array[col] > left_border_value * 0.9) {
            left_border_value = statistical_border_lr_array[col];
            left_border       = col;
        }

    // search for the right border
    uint16_t right_border       = srcPicture->width - 2;
    uint16_t right_border_value = 0;
    for (uint16_t col = srcPicture->width - 2; col > srcPicture->width * 0.83; col--)
        if(statistical_border_lr_array[col] > right_border_value * 0.9) {
            right_border_value = statistical_border_lr_array[col];
            right_border       = col;
        }


    // update location and size of field
    field_location_size[0] = left_border;       // left
    field_location_size[1] = right_border;      // right
}


/**
 * detect the border of the field
 * @param srcPicture the image
 * @return true, if field was found
 */
bool detectTableBorder(pictureDataP srcPicture) {
    // find potential field borders
    updateStatisticTB(srcPicture);
    updateStatisticLR(srcPicture);

    // analyse the potential field borders
    detectTableBorderTB(srcPicture);
    detectTableBorderLR(srcPicture);


    // compare the ratio of the real (120cm x 68cm) and the found field
    if (abs(((field_location_size[1] - field_location_size[0]) / (field_location_size[3] - field_location_size[2])) - (120.0 / 68.0)) < 0.1) {
        // found field - probably :)
        if(DRAW_FIELD) {
            uint16_t drawLeft   = field_location_size[0];
            uint16_t drawRight  = field_location_size[1];
            uint16_t drawTop    = field_location_size[2];
            uint16_t drawBottom = field_location_size[3];

            if (drawLeft == 0)
                drawLeft = 1;
            if (drawRight == 0)
                drawRight = srcPicture->width - 1;
            if (drawTop == 0)
                drawTop = 1;
            if (drawBottom == 0)
                drawBottom = srcPicture->height - 1;

            for (int8_t width = -1; width <= 1; width++) {
                drawLineVertical(  srcPicture, color_field, drawLeft   + width);
                drawLineVertical(  srcPicture, color_field, drawRight  + width);
                drawLineHorizontal(srcPicture, color_field, drawTop    + width);
                drawLineHorizontal(srcPicture, color_field, drawBottom + width);
            }
        }

        return true;
    } else {
        // adjust parameter
        return false;
    }
}


// the comment is located in the header file
void initializeDetect(pictureDataP srcPicture) {
    // allocate more than used - acceptable

    if (statistical_border_tb_array == NULL)
        statistical_border_tb_array = (uint64_t *) calloc(srcPicture->height, sizeof(uint64_t));

    if (statistical_border_lr_array == NULL)
        statistical_border_lr_array = (uint64_t *) calloc(srcPicture->width,  sizeof(uint64_t));
}

void finalizeDetect() {
    if (statistical_border_tb_array != NULL) {
        free(statistical_border_tb_array);
        statistical_border_tb_array = NULL;
    }

    if (statistical_border_lr_array != NULL) {
        free(statistical_border_lr_array);
        statistical_border_lr_array = NULL;
    }
}


// the comment is located in the header file
void processFrame(pictureDataP img) {
    // gaussianBlur2RGBA(img, img);

    if (detectTableBorder(img))
        detectCircle(img, 8, 12);
}
