#include "ffmpegio.h"
#include "pngedgedetect.h"
#include <libavutil/timestamp.h>
#include <libswscale/swscale.h>
#include <libavdevice/avdevice.h>

sig_atomic_t frames     = 0;

/**
 * This function must be called before anything else. Here we register
 * Codecs, Devices and Network stuff. This will ensure that we can
 * open every input/output that is supported by ffmpeg
 */
void ffmpeg_initialize()
{
    av_register_all();
    avformat_network_init();
    avdevice_register_all();
    av_log_set_level(AV_LOG_QUIET);
}

/**
 * Allocates Picture Structs needed for reading/writing
 * @param video
 * @param pix_fmt
 */
static void ffmpeg_alloc_av_structs(ffmpegIOS *video,
                                    enum AVPixelFormat pix_fmt)
{
    video->frame_count = 0;
    video->rescaled = false;
    av_init_packet(&(video->cur_packet));
    video->cur_frame = av_frame_alloc();
    video->cur_frame->width = video->width;
    video->cur_frame->height = video->height;
    avpicture_alloc(&(video->conv_frame),pix_fmt,
                    video->width,video->height);
}

/**
 * Allocates Picture Buffer
 * @param video
 */
static void ffmpeg_alloc_output_img(ffmpegIOS *video)
{
    avpicture_alloc(&(video->src_frame),
                    AV_PIX_FMT_YUV420P,
                    video->width,video->height);
    for(uint8_t i = 0; i < 8 ; i++)
    {
        video->cur_frame->data[i] = video->src_frame.data[i];
        video->cur_frame->linesize[i] = video->src_frame.linesize[i];
    }
}

/**
 * Mappes AVFrame to PictureData structure
 * @param video
 */
static void ffmpeg_alloc_image_struct(ffmpegIOS *video)
{
    video->image_info.width = video->width;
    video->image_info.height= video->height;
    video->image_info.color = COLOR_RGBA;
    video->image_info.raw   = video->conv_frame.data[0];
    video->image_info.rowPointers = calloc( video->height,
                                            sizeof(uint8_t **));
    for(uint16_t i = 0 ; i < video->height; i++)
        video->image_info.rowPointers[i] = 
            &(video->image_info.raw[video->width*i*4]);
}

/**
 * @param oc
 * @param codec
 * @param c
 * @param st
 * @param codec_id
 * @param frame_rate
 * @param width
 * @param height
 */
static void ffmpeg_add_stream(AVFormatContext *oc,
                              AVCodec **codec,
                              AVCodecContext **c,
                              AVStream **st,
                              enum AVCodecID codec_id,
                              uint16_t frame_rate,
                              uint16_t width,
                              uint16_t height)
{
    /* Resolution of video should be allways a multiple of 2 */
    if(width % 2 == 1 || height % 2 ==1)
    {
        soccerLog( "Resolution of Output Video must allways"
                        "be a multiple of 2");
    }
    /* find the encoder */

    *codec = avcodec_find_encoder(codec_id);
    if (!(*codec)) {
        soccerLog( "Could not find encoder for '%s'\n",
        avcodec_get_name(codec_id));
        exit(1);
    }
    *st = avformat_new_stream(oc, *codec);
    if (!st) {
        soccerLog( "Could not allocate stream\n");
        exit(1);
    }
    (*st)->id = oc->nb_streams-1;
    (*c) = (*st)->codec;
    (*c)->codec_id = codec_id;
    (*c)->height = height;
    (*c)->width = width;
    (*c)->bit_rate = 4*1024*1024;
    (*c)->bit_rate_tolerance = 4*1024*1024;
    (*c)->time_base.den = frame_rate;
    (*c)->time_base.num = 1;
    (*c)->gop_size = 12; /* emit one intra frame every twelve frames at most */
    (*c)->pix_fmt = AV_PIX_FMT_YUV420P;
    (*st)->time_base.num = 1;
    (*st)->time_base.den = frame_rate;
    /* Some formats want stream headers to be separate. */
    if (oc->oformat->flags & AVFMT_GLOBALHEADER)
        (*c)->flags |= CODEC_FLAG_GLOBAL_HEADER;
    return;
}

/**
 * Opens annd loads needed codecs for decoding the input file
 * @param filename The video file that will be decoded
 * @returns Struct that will be used to obtain frames of the video file
 */
ffmpegIOS *ffmpeg_open_video_input(const char *filename)
{
    ffmpegIOS* stream = calloc(1,sizeof(ffmpegIOS));
    int ret;

    if ((ret = avformat_open_input(&stream->fmt_ctx, filename, NULL, NULL)) < 0) {
        av_log(NULL, AV_LOG_ERROR, "Cannot open input %s \n%s\n",
            filename,av_err2str(ret));
        return NULL;
    }
    
    /* Analyse Input Video */
    
    if(avformat_find_stream_info(stream->fmt_ctx,NULL))
        soccerLog("Whoops");
    
    /* Finds best stream for Video Decoding. We Ignore the Audio stream
     * Because we'll only do Image recognition 
     */
    ret = av_find_best_stream(  stream->fmt_ctx, AVMEDIA_TYPE_VIDEO, -1,
                                -1, &stream->video_codec, 0);
    if (ret < 0) {
        av_log(NULL, AV_LOG_ERROR, "Cannot find a video stream "
            " in the input file\n");
        return NULL;
    }
    
    stream->stream_index = ret;
    stream->codec_ctx = 
        stream->fmt_ctx->streams[stream->stream_index]->codec;
    stream->width = stream->codec_ctx->width;
    stream->height = stream->codec_ctx->height;

    /* Initialize Decoder */
    AVDictionary * codec_options = ( 0 );
    av_dict_set(&codec_options, "threads","1",0);
    if ((ret = avcodec_open2(stream->codec_ctx, stream->video_codec, &codec_options)) < 0) {
        av_log(NULL, AV_LOG_ERROR, "Cannot open video decoder\n");
        return NULL;
    }
    stream->sws_ctx = sws_getContext(stream->width, 
                                     stream->height,
                                     stream->codec_ctx->pix_fmt, 
                                     stream->width, 
                                     stream->height,
                                     AV_PIX_FMT_RGBA, 0, 0, 0, 0);
    stream->eof = FFMPEG_EOF_NOTREACHED;
    ffmpeg_alloc_av_structs(stream, AV_PIX_FMT_RGBA);
    ffmpeg_alloc_image_struct(stream);
    return stream;
}

/**
 * Loads needed encoding codecs for writing into destination File
 * @param filename The location where the output file will be created
 * @param width The width of the video picture. Must be a multiple of 2
 * @param height The height of the video pciture. Must be a multiple of 2
 * @param frame_rate Specifies how many frames per second will be encoded
 * @returns ffmpeg io struct with all internal needed information
 *          for writing
 * 
 */
ffmpegIOS *ffmpeg_open_video_output(const char *filename, uint16_t width,
                                     uint16_t height, uint16_t frame_rate)
{
    ffmpegIOS * video = calloc(1,sizeof(ffmpegIOS));
    uint32_t ret =0 ;
    
    avformat_alloc_output_context2(&(video->fmt_ctx), NULL, "flv", filename);
    ffmpeg_add_stream(video->fmt_ctx,
                      &(video->video_codec),
                      &(video->codec_ctx),
                      &(video->video_stream),
                      CODEC_ID_H264,
                      frame_rate,
                      width,
                      height);
               
    AVDictionary * codec_options = ( 0 );
    av_dict_set(&codec_options, "threads","4",0);
    av_dict_set(&codec_options, "preset","ultrafast",0);
    avcodec_open2(video->codec_ctx,video->video_codec,&codec_options);
    /* Open Output file if needed */
    if (!(video->fmt_ctx->oformat->flags & AVFMT_NOFILE)) {
        ret = avio_open(&(video->fmt_ctx->pb), filename, AVIO_FLAG_WRITE);
        if (ret < 0) {
            soccerLog( "Could not open '%s': %s", filename,
            av_err2str(ret));
            return NULL;
        }
    }
    avformat_write_header(video->fmt_ctx,NULL);
    video->width = width;
    video->height = height;
    video->sws_ctx = sws_getContext(width, 
                                     height,
                                     AV_PIX_FMT_RGBA, 
                                     width, 
                                     height,
                                     AV_PIX_FMT_YUV420P, 0, 0, 0, 0);
    ffmpeg_alloc_av_structs(video, AV_PIX_FMT_RGBA);
    ffmpeg_alloc_image_struct(video);
    ffmpeg_alloc_output_img(video);
    video->eof = FFMPEG_EOF_UNKNOWN;
    
    video->cur_frame->format = AV_PIX_FMT_YUV420P;
    return video;
}

/**
 * 
 */
static void ffmpeg_log_packet(const AVFormatContext *pkt_ctx, const AVPacket *pkt)
{
    AVRational *time_base = &pkt_ctx->streams[pkt->stream_index]->time_base;
    printf("pts:%s pts_time:%s dts:%s dts_time:%s duration:%s "
        " duration_time:%s stream_index:%d\n",
        av_ts2str(pkt->pts), av_ts2timestr(pkt->pts, time_base),
        av_ts2str(pkt->dts), av_ts2timestr(pkt->dts, time_base),
        av_ts2str(pkt->duration), av_ts2timestr(pkt->duration, time_base),
        pkt->stream_index);
}

int32_t ffmpeg_seek_to_frame(ffmpegIOS *video, uint32_t frame)
{
    int32_t ret = 0;
    if(video->frame_count < frame)
    {
        while(video->frame_count < frame &&
              video->eof != FFMPEG_EOF_REACHED)
        {
            ffmpeg_get_next_frame(video);
        }
    }
}

/**
 * This will rescale the frame buffer of video with its rescale context
 * 
 */
void ffmpeg_rescale_frame(ffmpegIOS *video)
{
    if(!video->rescaled)
    {
        sws_scale(video->sws_ctx, 
                    (const uint8_t * const *)video->conv_frame.data,
                    video->conv_frame.linesize,
                    0,
                    video->height,
                    video->cur_frame->data,
                    video->cur_frame->linesize);
        video->rescaled = true;
    }
}

/**
 * 
 */
int32_t ffmpeg_get_next_frame(ffmpegIOS *video)
{
    int32_t ret = 0;
    int32_t got_frame = 0;
    
    if(video->eof == FFMPEG_EOF_REACHED)
    {
        return FFMPEG_EOF_REACHED;
    }
    av_init_packet(&video->cur_packet);
    do 
    {
        ret = av_read_frame(video->fmt_ctx, &video->cur_packet);
        if (ret >= 0 && video->cur_packet.stream_index == video->stream_index)
        {
            ret = avcodec_decode_video2(video->codec_ctx, 
                                        video->cur_frame, 
                                        &got_frame, 
                                        &video->cur_packet);
            if(ret < 0)
            {
                soccerLog(
                        "Something went wrong by decoding the video: %s",
                        av_err2str(ret));
                exit(ret);
            }   
        }
        else if(!(ret >= 0))
            return (video->eof = FFMPEG_EOF_REACHED);
    } while(got_frame<=0);
    av_free_packet(&video->cur_packet);
    sws_scale(video->sws_ctx, 
                        (const uint8_t * const *)video->cur_frame->data,
                        video->cur_frame->linesize,
                        0,
                        video->height,
                        video->conv_frame.data,
                        video->conv_frame.linesize);

    
    video->frame_count++;
    
    return video->eof;
}

/**
 *
 */
static inline int32_t ffmpeg_write_frame_helper(AVFormatContext *fmt_ctx,
                                                const AVRational *time_base,
                                                AVStream *st, AVPacket *pkt)
{
    /* rescale output packet timestamp values from codec to stream timebase */
    pkt->pts = av_rescale_q_rnd(pkt->pts, *time_base, st->time_base, 
        AV_ROUND_NEAR_INF|AV_ROUND_PASS_MINMAX);
    pkt->dts = av_rescale_q_rnd(pkt->dts, *time_base, st->time_base, 
        AV_ROUND_NEAR_INF|AV_ROUND_PASS_MINMAX);
    pkt->duration = av_rescale_q(fmt_ctx->duration, *time_base, st->time_base);
    pkt->stream_index = st->index;
    /* Write the compressed frame to the media file. */
    //ffmpeg_log_packet(fmt_ctx, pkt);
    return av_interleaved_write_frame(fmt_ctx, pkt);
}




/**
 * @param the ffmpeg context where we will write the buffer into the
 * stream
 * @returns EXIT_SUCCESS on success or kill program on error
 */
int32_t ffmpeg_write_frame(ffmpegIOS *video)
{
    int32_t ret = 0;
    int32_t got_packet = 0;
    
    //ffmpeg_rescale_frame(video);
    
    video->cur_frame->pts = video->frame_count;
        video->cur_frame->key_frame =
        video->cur_frame->pict_type =
            (video->frame_count%video->codec_ctx->gop_size==0) ? 
                AV_PICTURE_TYPE_I : 0;
    frames++;
    ret = avcodec_encode_video2(video->codec_ctx, &video->cur_packet,
                                video->cur_frame, &got_packet);
    if (ret < 0) {
        soccerLog("Error encoding video frame: %s", av_err2str(ret));
        exit(1);
    }
    video->frame_count++;
    if (got_packet)
        ret = ffmpeg_write_frame_helper(video->fmt_ctx,
                                        &video->codec_ctx->time_base,
                                        video->video_stream,
                                        &video->cur_packet);
    
    video->rescaled = false;
    
    return EXIT_SUCCESS;
}

/**
 * Frees all allocated Heap Memory of the ffmpeg io Struct
 * Calls needed specific free of libav library
 * @param video the ffmpegIOS struct that will be freed
 */
int32_t ffmpeg_close_video(ffmpegIOS *video)
{
    avio_close(video->fmt_ctx->pb);
    avcodec_close(video->codec_ctx);
    av_frame_free(&video->cur_frame);
    avformat_free_context(video->fmt_ctx);
    avpicture_free(&video->src_frame);
    avpicture_free(&video->conv_frame);
    av_free_packet(&video->cur_packet);
    sws_freeContext(video->sws_ctx);
    free(video->image_info.rowPointers);
    free(video);
    
    return EXIT_SUCCESS;
}
