#define _GNU_SOURCE
#include "mpicomm.h"
#include "pngedgedetect.h"
#include "pipeline.h"
#include <unistd.h>
#include <omp.h>

int32_t          myRank     = 0;
int32_t          worldSize  = 0;
uint64_t         frame = 0; /**< Global for package generation */
uint64_t         job_recv_buffer[2];
MpiJobRequest *  jobrequestSend;
MpiJobRequest ** jobrequestRecv;
MpiJob        ** jobs;

MPI_Request      workerJobRequest  = MPI_REQUEST_NULL,
                 workerrecvRequest = MPI_REQUEST_NULL,
                 workerYuvSend     = MPI_REQUEST_NULL,
              *  mainSendRequest,
              *  mainRecvRequest,
              *  mainYuvRecv;

MpiNode       *  nodes;
bool             multiNodeMode = false;

size_t           sizeOfYuvPackage = 0;
size_t           sizeOfYuvFrame   = 0;
size_t           sizeOfYuvHeader  = sizeof(uint64_t);

uint8_t       *  yuvSendBuffer;
uint8_t       ** yuvRecvBuffer;


bool isMpiMainNode()
{
    return (myRank == 0);
}

int32_t getMpiRank()
{
    return myRank;
}

int32_t getWorldSize()
{
    return worldSize;
}

void * mpi_comm_alloc(size_t bytes)
{
    void * ptr = calloc(bytes,1);
    //soccerLog("[0x%x] Allocated %d Bytes",ptr,bytes);
    return ptr;
}

void createMpiDataTypes()
{
    // JobRequest
    const int       jrItems = 4;
          int       jrBlockLengths[4] = {1,1,1,1};
    MPI_Datatype    jrtypes[4]        = {MPI_UINT64_T, MPI_UINT64_T,
                                         MPI_UINT64_T, MPI_UINT64_T};
    MPI_Aint        jroffset[4];
    
    jroffset[0] = offsetof(MpiJobRequest, rank);     
    jroffset[1] = offsetof(MpiJobRequest, num_of_threads); 
    jroffset[2] = offsetof(MpiJobRequest, last_frame_read); 
    jroffset[3] = offsetof(MpiJobRequest, last_frame_writen); 
    
    MPI_Type_create_struct( jrItems, 
                            jrBlockLengths, 
                            jroffset, 
                            jrtypes, 
                            &MPI_JOB_REQUEST_T);
    MPI_Type_commit(&MPI_JOB_REQUEST_T);
                            
    const int       jpItems             = 2;
          int       jpBlockLengths[2]   = {1,1};
    MPI_Datatype    jpTypes[2]          = {MPI_UINT64_T, MPI_UINT64_T};
    MPI_Aint        jpoffset[2];
    
    jpoffset[0] = offsetof(MpiJob, first_frame);
    jpoffset[1] = offsetof(MpiJob, last_frame);
    
    MPI_Type_create_struct( jpItems, 
                            jpBlockLengths, 
                            jpoffset, 
                            jpTypes, 
                            &MPI_JOB_PACKAGE_T);
    MPI_Type_commit(&MPI_JOB_PACKAGE_T);
}


void prepareJobRequest(MpiJobRequest * jobrequest)
{
    jobrequest->rank = (uint64_t) myRank;
    jobrequest->last_frame_read = packetsIn * pipeline_get_max_size();
    jobrequest->last_frame_writen = packetsOut * pipeline_get_max_size();
    jobrequest->num_of_threads = (uint64_t) omp_get_max_threads();
}

void prepareJobPacket(MpiJob * newjob)
{
    newjob->first_frame = frame;
    newjob->last_frame = frame+ pipeline_get_max_size();
    frame += pipeline_get_max_size();
}

void checkJobRequestSend(   MPI_Request mpiOut,
                            MPI_Request mpiIn, 
                            MpiJobRequest * jobr,
                            GAsyncQueue * jobQueue)
{
    int readyflag = false;
    MPI_Test(&mpiOut,&readyflag,NULL);
    // To ensure we don't make to much jobrequests and all other nodes
    // have to wait for us. We enforce that our queue should only
    // contain 10 Packages 
    if(         ffmpeg_io_input->eof != FFMPEG_EOF_REACHED &&
                g_async_queue_length(jobQueue) < 2 &&
                readyflag)
    {
        prepareJobRequest(jobr);
        MPI_Start(&workerJobRequest);
    }
}

void checkJobPackageRecv(MPI_Request mpiRecv,
                         GAsyncQueue * jobQueue)
{
    int readyflag = false;
    MPI_Test(&mpiRecv,&readyflag,NULL);
    if(readyflag)
    {
        MpiJobP tmp = mpi_comm_alloc(sizeof(MpiJob));
        tmp->first_frame = job_recv_buffer[0];
        tmp->last_frame = job_recv_buffer[1];        
        g_async_queue_push(jobQueue,tmp);
        MPI_Start(&workerrecvRequest);
    }
}

void logJobPackage(MpiJob ** jobs, int i)
{
    FILE * fp = fopen("job.send","a");
    
    fprintf(fp, "%ld ; %d ; %ld ; %ld\n",
        time(NULL),i,jobs[i]->first_frame,jobs[i]->last_frame);
    
    fclose(fp);
}

void logFramePackage(int first_frame, int i)
{
    FILE * fp = fopen("job.recv","a");
    
    fprintf(fp, "%ld ; %d ; %d ; %d\n",time(NULL),i,first_frame,first_frame+ pipeline_get_max_size());

    fclose(fp);
}

void checkJobPackageSend(   MPI_Request mpiOut[],
                            MPI_Request mpiIn[], 
                            MpiJob ** jobs,
                            MpiJobRequest ** jobrequests)
{
    for(int i = 0 ; i < worldSize; i++)
    {
        int readySend = false;
        int readyRecv = false;
        MPI_Test(&mpiOut[i],&readySend,NULL);
        if( readySend &&
            MPI_Test(&mpiIn[i],&readyRecv,NULL) == MPI_SUCCESS &&
            readyRecv &&
            nodes[i].jobsSend - (queue_max_size * omp_get_max_threads()) < nodes[i].jobsRecv)
        {
            prepareJobPacket(jobs[i]);
            
            logJobPackage(jobs, i);
            nodes[i].jobsSend++;
            
            MPI_Start(&mpiOut[i]);
            MPI_Start(&mpiIn[i]);;
        }
    }
}

    
void initializeMPI(int argc, char * argv[]) 
{
    MPI_Init(&argc, &argv);
    MPI_Comm_rank(MPI_COMM_WORLD, &myRank);
    MPI_Comm_size(MPI_COMM_WORLD, &worldSize);

    createMpiDataTypes();

    if(worldSize <= 1)
        soccerLog("===Single Node Mode===");
    else
        if(myRank == 0) {
            multiNodeMode = true; // todo: korrekt hier?
            soccerLog("===Multi Node Mode===");
        } 
        else
            soccerLog("===Hello===");
}

void allocateMpiMemory()
{
    /* Memory Allocation and Initialization */
    
    jobrequestSend  = mpi_comm_alloc(sizeof(MpiJobRequest));
    if(isMpiMainNode())
    {
        nodes           = mpi_comm_alloc(worldSize * sizeof(MpiNode));
        yuvRecvBuffer   = (uint8_t **) mpi_comm_alloc(worldSize * sizeof(uint8_t *));
        mainSendRequest = mpi_comm_alloc(worldSize * sizeof(MPI_Request));
        mainRecvRequest = mpi_comm_alloc(worldSize * sizeof(MPI_Request));
        mainYuvRecv     = mpi_comm_alloc(worldSize * sizeof(MPI_Request));
        jobrequestRecv  = mpi_comm_alloc(worldSize * sizeof(MpiJobRequest *));
        jobs            = mpi_comm_alloc(worldSize * sizeof(MpiJob *));
        for(int i = 0 ; i < worldSize; i++)
        {
            mainSendRequest[i] = MPI_REQUEST_NULL;
            mainRecvRequest[i] = MPI_REQUEST_NULL;
            yuvRecvBuffer[i]    = mpi_comm_alloc(sizeOfYuvPackage * sizeof(uint8_t));
            jobs[i]             = mpi_comm_alloc(sizeof(MpiJob));
            jobrequestRecv[i]   = mpi_comm_alloc(sizeof(MpiJobRequest));
        }  
    }
    else
    {
        yuvSendBuffer   = mpi_comm_alloc(sizeOfYuvPackage);
    }  
     
}

pipelinePacketP deserializePipelinePacket(uint16_t node)
{
    pipelinePacketP tmp = pipeline_alloc_package();
    // Restore Header Information
    memcpy(&(tmp->id),yuvRecvBuffer[node],sizeOfYuvHeader);
    for(int i = 0; i < pipeline_get_max_size(); i++)
    {
        tmp->items[i].img = image_new_only_avPicture(ffmpeg_io_input->width,
                                                     ffmpeg_io_input->height,
                                                     AV_PIX_FMT_YUV420P);
        memcpy(tmp->items[i].img->yuvFrame.data[0],
               yuvRecvBuffer[node] + sizeOfYuvHeader + (i * sizeOfYuvFrame),
               sizeOfYuvFrame);
        tmp->items[i].ready = true;
        tmp->items[i].id = tmp->id + i;
    }
    //logFramePackage(tmp->id,node);
    tmp->ready = true;
    return tmp;
}

void serializePipelinePacket(pipelinePacketP pkg)
{
    // Write Header Information
    memcpy(yuvSendBuffer,&(pkg->id),sizeOfYuvHeader);
    for(int i = 0; i < pipeline_get_max_size(); i++)
    {
        if(pkg->items[i].img != NULL)
        {
            memcpy(yuvSendBuffer + sizeOfYuvHeader + (i * sizeOfYuvFrame),
                   pkg->items[i].img->yuvFrame.data[0],
                   sizeOfYuvFrame);
        }
        else
        {
            memset(yuvSendBuffer + sizeOfYuvHeader + (i * sizeOfYuvFrame),
                0,sizeOfYuvFrame);
        }
    }    
    pipeline_free_package(pkg);
}

void recvPipelinePacket(GAsyncQueue * queue)
{
    for(int i = 0 ; i < worldSize; i++)
    {
        int readyflag = false;
        MPI_Test(&mainYuvRecv[i],&readyflag,NULL);
        if(readyflag)
        {
            packetsIn++;
            nodes[i].jobsRecv++;
            g_async_queue_push(queue,deserializePipelinePacket(i));
            MPI_Start(&mainYuvRecv[i]);
        }
    }
}

void increaseMainCounterBy(int i)
{
    if(nodes != NULL)
        nodes[0].jobsRecv += i;
}

void sendPipelinePacket(GAsyncQueue * queue)
{
    int readyflag = false;
    MPI_Test(&workerYuvSend,&readyflag,NULL);
    if(readyflag && g_async_queue_length(queue) > 0)
    {
        pipelinePacketP pkg = g_async_queue_timeout_pop(queue, 200000);
        if(pkg != NULL)
        {
            serializePipelinePacket(pkg);
            MPI_Start(&workerYuvSend);
        }
    }
}

void startInitialMpiRequests()
{
    if(isMpiMainNode())
    {
        for(int i = 0; i < worldSize; i++)
        {
            MPI_Send_init(jobs[i],2,MPI_UINT64_T,
                          i,MPI_JOB_PACKAGE, MPI_COMM_WORLD,
                          &mainSendRequest[i]);
            MPI_Recv_init(jobrequestRecv[i],4,MPI_UINT64_T,
                          i,MPI_JOB_REQUEST, MPI_COMM_WORLD,
                          &mainRecvRequest[i]);
            MPI_Recv_init(yuvRecvBuffer[i],sizeOfYuvPackage, MPI_UINT8_T,
                          i,MPI_YUV_PACKAGE, MPI_COMM_WORLD,
                          &mainYuvRecv[i]);
            MPI_Start(&mainRecvRequest[i]);
            MPI_Start(&mainYuvRecv[i]);
        }
    }
    else
    {
        MPI_Send_init(yuvSendBuffer,sizeOfYuvPackage,MPI_UINT8_T,
                  MPI_MAIN_NODE,MPI_YUV_PACKAGE, MPI_COMM_WORLD,
                  &workerYuvSend);
    }
    MPI_Send_init(jobrequestSend,1,MPI_JOB_REQUEST_T,
                  MPI_MAIN_NODE,MPI_JOB_REQUEST, MPI_COMM_WORLD,
                  &workerJobRequest);
    MPI_Recv_init(job_recv_buffer,1,MPI_JOB_PACKAGE_T,
                  MPI_MAIN_NODE,MPI_JOB_PACKAGE, MPI_COMM_WORLD,
                  &workerrecvRequest);
    // We Start the MPI_Recv Request 
    MPI_Start(&workerrecvRequest);
}

void logNodes()
{
    FILE * fp = fopen("job.nodes","a");
    
    fprintf(fp,"%ld ; ", time(NULL));
    for(int i = 0; i < worldSize ; i++)
    {
        fprintf(fp,"%ld ; %ld ;",nodes[i].jobsSend,nodes[i].jobsRecv);
    }
    fprintf(fp,"\n");
    fclose(fp);
}

void mpiComm(GAsyncQueue * jobQueue,
             GAsyncQueue * yuvIO,
             uint16_t width, 
             uint16_t heigth,
             size_t   bpp)
{
    sizeOfYuvFrame = (width * heigth * 1.5);
    sizeOfYuvPackage =  sizeOfYuvFrame * pipeline_get_max_size()
                      + sizeOfYuvHeader;   
    allocateMpiMemory();
    startInitialMpiRequests();
    createMpiDataTypes();
    while(  packetprocessor > 0 ||
            !queueFillerExited ||
            packetsIn != packetsOut ||
            (isMpiMainNode() && (ffmpeg_io_input->frame_count == ffmpeg_io_output->frame_count)) ||
            (!isMpiMainNode() && g_async_queue_length(yuvIO)) > 0)
    {
        checkJobPackageRecv(workerrecvRequest,
                            jobQueue);
        checkJobRequestSend(workerJobRequest, 
                            workerrecvRequest, 
                            jobrequestSend,
                            jobQueue);
       
        if(isMpiMainNode())
        {
            checkJobPackageSend(mainSendRequest,mainRecvRequest,
                                jobs,jobrequestRecv);
            recvPipelinePacket(yuvIO);
            logNodes();
        }
        else
        {
            sendPipelinePacket(yuvIO);
        }
       
        usleep(10);
    }
    
    soccerLog("MPI DONE");
}

void freeMPI()
{
    MPI_Finalize();
    free(nodes);
}
