#include "pipeline.h"
#include <libswscale/swscale.h>

uint8_t pipelineSize = 4;

void pipeline_set_max_size(uint8_t size) {
    pipelineSize = size;
}

uint8_t pipeline_get_max_size() {
    return pipelineSize;
}

/**
 * Allocates buffers for a packet in a memory compact way. For better
 * processor caching
 * @returns a pointer to a allocated packet
 */
pipelinePacketP pipeline_alloc_package() {
    uint64_t memsize    = pipelineSize * (sizeof(pipelineItem)) + sizeof(pipelinePacket);
    void * memptr       = calloc(memsize, 1);
    pipelinePacketP pkg = memptr;
    pkg->items          = (pipelineItemP) (memptr + sizeof(pipelinePacket));
    memptr             += pipelineSize * sizeof(pipelinePacketP);

    return pkg;
}

pipelinePacketP pipeline_fill_package(pipelinePacketP pkg, ffmpegIOSP input) {
    if(pkg == NULL)
        pkg = pipeline_alloc_package();

    pkg->id         = input->frame_count;
    pictureData tmpPicture = {0};
    AVPicture   tmpAVPicture   = {0};
    tmpPicture.raw         = input->conv_frame.data[0];
    tmpAVPicture.data[0]      = input->cur_frame->data[0];
    for(uint8_t i = 0 ; i < pipelineSize; i++) {
        if(pkg->items[i].img == NULL)
            pkg->items[i].img =
                    image_new(input->image_info.width,
                              input->image_info.height,
                              input->image_info.color);

        pkg->items[i].id          = input->frame_count;
        input->conv_frame.data[0] = pkg->items[i].img->raw;
        input->cur_frame->data[0] = pkg->items[i].img->yuvFrame.data[0];

        if(ffmpeg_get_next_frame(input) == FFMPEG_EOF_REACHED)
        {
            image_free(pkg->items[i].img);
            pkg->items[i].img = NULL;
            break;
        }
    }

    input->conv_frame.data[0] = tmpPicture.raw;
    input->cur_frame->data[0] = tmpAVPicture.data[0];
    
    return pkg;
}

void pipeline_free_package(pipelinePacketP pkg) {
    for(uint8_t i = 0; i < pipelineSize; i++)
        image_free(pkg->items[i].img);

    free(pkg);
}

int pipeline_foreach(pipelinePacketP pkg, void (*func)(pictureDataP)) {
    for(uint8_t i = 0; i < pipelineSize; i++)
    {
        if(pkg->items[i].img != NULL)
        {
            func(pkg->items[i].img);
            pkg->items[i].ready = true;

            image_alloc_avPicture(pkg->items[i].img, AV_PIX_FMT_YUV420P);
            sws_scale(pkg->rgb2yuv, 
                (const uint8_t * const*)&pkg->items[i].img->raw,
                pkg->items[i].img->lineSize,
                0,
                pkg->items[i].img->height,
                pkg->items[i].img->yuvFrame.data,
                pkg->items[i].img->yuvFrame.linesize);
            image_free_raw(pkg->items[i].img);
        }
    }

    pkg->ready = true;

    return EXIT_SUCCESS;
}

pipelineStatus pipeline_write_packet(pipelinePacketP pkg, ffmpegIOSP output) {
    if(!pkg->ready)
        return PACKET_NOT_READY;

    for(uint8_t i = 0; i < pipelineSize; i++)
        if(!(pkg->items[i].ready))
            return PACKET_ITEM_NOT_READY;

    if(output->frame_count != pkg->id)
        return PACKET_FRAME_MISMATCH;
    
    for(uint8_t i = 0; i < pipelineSize; i++)
        pipeline_write_item(&pkg->items[i], output);

    return PACKET_SUCESSFULLY_WRITTEN;
}

pipelineStatus pipeline_write_item(pipelineItemP item, ffmpegIOSP output) {
    if(!item->ready)
        return PACKET_ITEM_NOT_READY;

    if(output->frame_count != item->id)
        return PACKET_FRAME_MISMATCH;
        
    if(item->img == NULL)
        return PACKET_ITEM_NOT_READY;

    AVPicture tmpAV = {0};
    
    memcpy(tmpAV.data,output->cur_frame->data, 8*sizeof(void*));
    memcpy(output->cur_frame->data,&item->img->yuvFrame.data,8*sizeof(void *));
    ffmpeg_write_frame(output);
    memcpy(output->cur_frame->data,tmpAV.data,8*sizeof(void *));

    return PACKET_SUCESSFULLY_WRITTEN;
}
