
#include "ffmpegio.h"
#include <stdint.h>
#include <stdbool.h>

#ifndef PIPELINE_H
#define PIPELINE_H


typedef enum pipelinestatus{
    PACKET_SUCESSFULLY_WRITTEN  = 0, /**<  */
    PACKET_NOT_READY            = 1 << 0, /**<  */
    PACKET_ITEM_NOT_READY       = 1 << 1, /**<  */
    PACKET_FRAME_MISMATCH       = 1 << 2 /**<  */
} pipelineStatus;


/**
 * @brief Describes a Pipeline Frame
 * 
 * 
 */
typedef struct pipelineitem {
    uint64_t        id; /**< unique frame id */
    pictureDataP    img; /**< pointer with frame information */
    bool            ready; /**< this should only be true if processing this frame was sucessful */
} pipelineItem;

typedef pipelineItem * pipelineItemP;


/**
 * @brief Describes multiple Pipeline items as a package
 * 
 * 
 */
typedef struct pipelinepacket {
    uint64_t        id; /**< first frame id of the packet */
    pipelineItemP   items; /**< Array of pipeline items */
    bool            ready; /**< true if all items are processed properly */
    struct SwsContext * rgb2yuv; /**< pointer to a color conversion from rgba to yuv420p */
    struct SwsContext * yuv2rgb; /**< pointer to a color conversion from yuvxxx to rgba */
} pipelinePacket;

typedef pipelinePacket *  pipelinePacketP;
typedef pipelinePacketP * pipelinePacketPP;


/**
 * Sets the actual maximum size for a pipeline.
 * This should be only called before any pipeline 
 * actions have been done. Calling this after packets 
 * already have been allocated will break the functions.
 * @param size the maximum size for a pipeline packet
 */
void pipeline_set_max_size(uint8_t size);

/**
 * @returns the actual maximum size for a pipeline
 */
uint8_t pipeline_get_max_size();

/**
 * Fills the pipeline Packet with pipeline_get_max_size
 * frames from the given ffmpegIO.
 * @param pkg pointer to an allocated packet or NULL
 * if this function should allocate the needed buffers.
 * @param input the source where we get the frames from.
 * @return pointer to packet struct. If pkg is provided we return (*pkg)
 */
pipelinePacketP pipeline_fill_package(pipelinePacketP pkg, ffmpegIOSP input);

/**
 * Frees all allocated buffers of a pkg
 * @param pkg the packet that will be freed
 * @return EXIT_SUCCESS on success
 */
void pipeline_free_package(pipelinePacketP pkg);

/**
 * Takes a packet and calls for each item the given function.
 * @param okg a packet allocated and filled from pipeline_fill_package.
 * @param *func a function that takes pictureData and manipulates the image.
 * @return EXIT_SUCCESS on sucess
 */
int pipeline_foreach(pipelinePacketP pkg, void (*func)(pictureDataP));

/**
 * Writes buffers of a pipeline into a ffmpeg stream. Only if every item in the 
 * packet is ready and the timestamp of the packet matches with the stream timestamp.
 * @param pkg the packet that will be flushed into the output.
 * @param output ffmpeg output struct that will be written in.
 * @returns PACKET_NOT_READY if ready flag of packet was not set.
 * PACKET_ITEM_NOT_READY if one of the items ready flag was not set.
 * PACKET_FRAME_MISMATCH if the timestamp of packet and output don't match.
 */
pipelineStatus pipeline_write_packet(pipelinePacketP pkg, ffmpegIOSP output);

/**
 * Writes a pipelineitem into a ffmpeg stream. Only if item is ready and 
 * the timestamp of the item matches with the stream timestamp.
 * @param item the item that will be flushed into the output.
 * @param output ffmpeg output struct that will be written in.
 * @returns PACKET_ITEM_NOT_READY if the item ready flag was not set
 * PACKET_FRAME_MISMATCH if the timestamp of the item and output don't match
 */
pipelineStatus pipeline_write_item(pipelineItemP item, ffmpegIOSP output);

pipelinePacketP pipeline_alloc_package();
#endif
